/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * gnome-print.c: Main GnomePrint API
 *
 * Authors:
 *   Raph Levien (raph@acm.org)
 *   Miguel de Icaza (miguel@kernel.org)
 *   Lauris Kaplinski <lauris@ariman.ee>
 */
/* Must include these two first */
#include "config.h"
#include <libgnomeprint/gnome-print-i18n.h>

#include <string.h>
#include <locale.h>
#include <unicode.h>

#include <gtk/gtk.h>
#include <stdio.h>
#include <libart_lgpl/art_affine.h>
#include <libart_lgpl/art_vpath.h>
#include <libart_lgpl/art_bpath.h>
/* gnome.h is only needed for g_concat_string_and_file */
#include <gnome.h>

#include <libgnomeprint/gnome-printer-private.h>
#include <libgnomeprint/gnome-print.h>
#include <libgnomeprint/gnome-print-private.h>

/* Gnome Print Drivres */
#include <libgnomeprint/gnome-print-ps.h>
#include <libgnomeprint/gnome-print-ps2.h>
#include <libgnomeprint/gnome-print-pdf.h>
#include <libgnomeprint/gnome-print-pclr.h>
#include <libgnomeprint/gnome-print-pclv.h>
#include <libgnomeprint/gnome-print-pixbuf.h>
#include <libgnomeprint/gnome-print-frgba.h>

#include <libgnomeprint/gnome-font-private.h>
#include <libgnomeprint/gnome-pgl.h>
#include <libgnomeprint/gnome-pgl-private.h>
#include <libgnomeprint/gp-ps-unicode.h>


static void gnome_print_context_class_init (GnomePrintContextClass *klass);

static void gnome_print_context_init (GnomePrintContext *pc);

static void gnome_print_context_finalize (GtkObject *object);

static GtkObjectClass *parent_class = NULL;

GtkType
gnome_print_context_get_type (void)
{
	static GtkType pc_type = 0;
	
	if (!pc_type)
	{
		GtkTypeInfo pc_info =
		{
			"GnomePrintContext",
			sizeof (GnomePrintContext),
			sizeof (GnomePrintContextClass),
			(GtkClassInitFunc) gnome_print_context_class_init,
			(GtkObjectInitFunc) gnome_print_context_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
		
		pc_type = gtk_type_unique (gtk_object_get_type (), &pc_info);
	}
	
	return pc_type;
}

static void
gnome_print_context_class_init (GnomePrintContextClass *class)
{
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) class;

	parent_class = gtk_type_class (gtk_object_get_type ());

	object_class->finalize = gnome_print_context_finalize;
#if 0
	/* Well - according to spec GtkObjectClass if zeroed */
	class->newpath = NULL;
	class->moveto = NULL;
	class->lineto = NULL;
	class->curveto = NULL;
	class->closepath = NULL;
	class->setrgbcolor = NULL;
	class->fill = NULL;
	class->eofill = NULL;
	class->setlinewidth = NULL;
	class->setmiterlimit = NULL;
	class->setlinejoin = NULL;
	class->setlinecap = NULL;
	class->setdash = NULL;
	class->strokepath = NULL;
	class->stroke = NULL;
	class->setfont = NULL;
	class->show_ucs4 = NULL;
	class->concat = NULL;
#ifdef SETMATRIX_DEPRECATED
	class->setmatrix = NULL;
#endif
	class->gsave = NULL;
	class->grestore = NULL;
	class->clip = NULL;
	class->eoclip = NULL;
	class->grayimage = NULL;
	class->rgbimage = NULL;
	class->textline = NULL;
	class->beginpage = NULL;
	class->showpage = NULL;
	class->close = NULL;
	class->setopacity = NULL;
	class->rgbaimage = NULL;
#endif
}

static void
gnome_print_context_init (GnomePrintContext *pc)
{
	pc->f = NULL;
}

GnomePrintContext *
gnome_print_context_new_with_paper_size (GnomePrinter *printer, const char *paper_size)
{
	g_return_val_if_fail (printer != NULL, NULL);
	g_return_val_if_fail (GNOME_IS_PRINTER (printer), NULL);
	g_return_val_if_fail (paper_size != NULL, NULL);

	/* This is just TEMPORARY !*/

	if (strcmp (printer->driver, "gnome-print-ps") == 0) {
		GnomePrintPs *ps = gnome_print_ps_new (printer);
		GnomePrintContext *frgba = gnome_print_frgba_new ((GnomePrintContext *) ps);
		return frgba ? frgba : NULL;
	}
	if (strcmp (printer->driver, "gnome-print-ps-rgb") == 0) {
		GnomePrintPs *ps = gnome_print_ps_new (printer);
		return ps ? GNOME_PRINT_CONTEXT (ps) : NULL;
	}
	if (strcmp (printer->driver, "gnome-print-ps2") == 0) {
		GnomePrintPs2 *ps2 = gnome_print_ps2_new (printer);
		GnomePrintContext *frgba = gnome_print_frgba_new ((GnomePrintContext *) ps2);
		return frgba ? frgba : NULL;
	}
	if (strcmp (printer->driver, "gnome-print-pdf") == 0) {
		GnomePrintPdf *pdf = gnome_print_pdf_new (printer);
		return pdf ? GNOME_PRINT_CONTEXT (pdf) : NULL;
	}

	if (strcmp (printer->driver, "gnome-print-pclr") == 0) {
		return gnome_print_pclr_new (printer, paper_size, 300);
	}
	if (strcmp (printer->driver, "gnome-print-pclv") == 0) {
		return gnome_print_pclv_new (printer, paper_size, 300);
	}

#if 0
	if (strncmp (printer->driver, "gnome-print-uni", strlen ("gnome-print-uni")) == 0)
		return gnome_print_pixbuf_new (printer, paper_size, 75);
#endif

	return NULL;
}

/*
 * This dispatch is by cases rather than being object oriented. It may
 * be desirable to make it more OO. Basically, the way that would be
 * done would be to have a gnome_printer_make_new_context () method
 * on GnomePrinter objects.
 * 
 * However, this will do for now, and saves making extra classes.
*/
GnomePrintContext *
gnome_print_context_new (GnomePrinter *printer)
{
	g_return_val_if_fail (printer != NULL, NULL);

	return gnome_print_context_new_with_paper_size (printer, "US-Letter");
}

static void
gnome_print_context_finalize (GtkObject *object)
{
	GnomePrintContext *pc;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GNOME_IS_PRINT_CONTEXT (object));

	pc = GNOME_PRINT_CONTEXT (object);

	(* GTK_OBJECT_CLASS (parent_class)->finalize) (object);
}

/* methods */

#define PRINT_CLASS(pc) GNOME_PRINT_CONTEXT_CLASS(GTK_OBJECT(pc)->klass)

int
gnome_print_newpath (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->newpath (pc);
}

int
gnome_print_moveto (GnomePrintContext *pc, double x, double y)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->moveto (pc, x, y);
}

int
gnome_print_lineto (GnomePrintContext *pc, double x, double y)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->lineto (pc, x, y);
}

int
gnome_print_curveto (GnomePrintContext *pc, double x1, double y1, double x2, double y2, double x3, double y3)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->curveto (pc, x1, y1, x2, y2, x3, y3);
}

int
gnome_print_closepath (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->closepath (pc);
}

int
gnome_print_setrgbcolor (GnomePrintContext *pc, double r, double g, double b)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setrgbcolor (pc, r, g, b);
}

int
gnome_print_fill (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->fill (pc);
}

int
gnome_print_eofill (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->eofill (pc);
}

int
gnome_print_setlinewidth (GnomePrintContext *pc, double width)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setlinewidth (pc, width);
}

int
gnome_print_setmiterlimit (GnomePrintContext *pc, double limit)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setmiterlimit (pc, limit);
}

int
gnome_print_setlinejoin (GnomePrintContext *pc, int jointype)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setlinejoin (pc, jointype);
}

int
gnome_print_setlinecap (GnomePrintContext *pc, int captype)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setlinecap (pc, captype);
}

int
gnome_print_setdash (GnomePrintContext *pc, int n_values, const double *values, double offset)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	if (n_values != 0)
		g_return_val_if_fail (values != NULL, -1);

	return PRINT_CLASS(pc)->setdash (pc, n_values, values, offset);
}

int
gnome_print_strokepath (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->strokepath (pc);
}

int
gnome_print_stroke (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->stroke (pc);
}

int
gnome_print_setfont (GnomePrintContext *pc, GnomeFont *font)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (font != NULL, -1);

	return PRINT_CLASS(pc)->setfont (pc, font);
}

int
gnome_print_show (GnomePrintContext *pc, char const *text)
{
	static gboolean uerror = FALSE;

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (text != NULL, -1);
	g_return_val_if_fail (!uerror, -1);

	if (strlen (text) < 1) return 0;

	if (PRINT_CLASS (pc)->show_ucs4) {
		static gboolean uinit = FALSE;
		static unicode_iconv_t uiconv = (unicode_iconv_t) -1;
		const gchar * ibuf;
		guint32 * buf;
		gchar * obuf;
		size_t ilen, olen;
		ssize_t converted;
		gint ret;

		if (!uinit) {
			unicode_init ();
			uiconv = unicode_iconv_open ("UCS4", "UTF-8");
			if (uiconv == (unicode_iconv_t) -1) {
				g_warning ("Could not initilize unicode convertion");
				uerror = TRUE;
				return -1;
			}
			uinit = TRUE;
		}

		ret = -1;

		ilen = strlen (text);
		ibuf = (const gchar *) text;
		olen = ilen * 2 * sizeof (guint32);
		buf = g_new (guint32, 2 * ilen);
		obuf = (gchar *) buf;
		converted = unicode_iconv (uiconv, &ibuf, &ilen, &obuf, &olen);

		if (converted >= 0) {
			gint length;
			length = (guint32 *) obuf - buf;
			ret = (* PRINT_CLASS (pc)->show_ucs4) (pc, buf, length);
		}
		g_free (buf);
		return ret;
	}

	return -1;
}

int
gnome_print_show_ucs4 (GnomePrintContext * pc, guint32 * buf, gint length)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (buf != NULL, -1);
	g_return_val_if_fail (length > 0, -1);

	if (PRINT_CLASS (pc)->show_ucs4)
		return (* PRINT_CLASS (pc)->show_ucs4) (pc, buf, length);

	return -1;
}

int
gnome_print_concat (GnomePrintContext *pc, const double matrix[6])
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (matrix != NULL, -1);

	return PRINT_CLASS(pc)->concat (pc, matrix);
}

#if 0
/**
 * gnome_print_concat_optimize:
 * @gpc: 
 * @: 
 * 
 * This method should be use when apps need to concat a matrix
 * that will be used on all of the page. For example a scaling factor or
 * landscape rotation+translation. No harm is done if this method is a) not
 * used or b) not used correctly. This function provides a mayor optimization
 * when scaling is to be done in the whole page and when the page needs to be
 * landscape rotated.
 **/
int
gnome_print_concat_optimize (GnomePrintContext *pc, const double matrix[6])
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (matrix != NULL, -1);

	g_warning ("DO NOT use this function. It is experimental and will be removed\n");
	
	if (PRINT_CLASS (pc)->concat_optimize != NULL)
		return PRINT_CLASS(pc)->concat_optimize (pc, matrix);
	else
		return PRINT_CLASS(pc)->concat (pc, matrix);
}
#endif

#ifdef SETMATRIX_DEPRECATED
int
gnome_print_setmatrix (GnomePrintContext *pc, const double matrix[6])
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (matrix != NULL, -1);

	return PRINT_CLASS(pc)->setmatrix (pc, matrix);
}
#endif

int
gnome_print_gsave (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->gsave (pc);
}

int
gnome_print_grestore (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->grestore (pc);
}

int
gnome_print_clip (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->clip (pc);
}

int
gnome_print_eoclip (GnomePrintContext *pc)
{
	return PRINT_CLASS(pc)->eoclip (pc);
}

int
gnome_print_grayimage (GnomePrintContext *pc, const char *data, int width, int height, int rowstride)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (data != NULL, -1);

	return PRINT_CLASS(pc)->grayimage (pc, data, width, height, rowstride);
}

int
gnome_print_rgbimage (GnomePrintContext *pc, const char *data, int width, int height, int rowstride)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (data != NULL, -1);

	return PRINT_CLASS(pc)->rgbimage (pc, data, width, height, rowstride);
}

int
gnome_print_rgbaimage (GnomePrintContext *pc, const char *data, int width, int height, int rowstride)
{
	guchar * b, * d;
	const guchar * s;
	gint x, y, alpha, tmp, ret;

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (data != NULL, -1);

	if (PRINT_CLASS (pc)->rgbaimage != NULL) {

		/* We have REAL rgba method */

		return PRINT_CLASS (pc)->rgbaimage (pc, data, width, height, rowstride);
	}

	if (PRINT_CLASS (pc)->rgbimage == NULL) return 0;

	/* Print RGB - better than nothing */

	b = g_new (gchar, width * height * 3);
	g_return_val_if_fail (b != NULL, -1);

	for (y = 0; y < height; y++) {
		s = data + y * rowstride;
		d = b + y * width * 3;
		for (x = 0; x < width; x++) {
			alpha = s[3];
			tmp = (s[0] - 0xff) * alpha;
			d[0] = 0xff + ((tmp + 0x80) >> 8);
			tmp = (s[1] - 0xff) * alpha;
			d[1] = 0xff + ((tmp + 0x80) >> 8);
			tmp = (s[2] - 0xff) * alpha;
			d[2] = 0xff + ((tmp + 0x80) >> 8);
			s += 4;
			d += 3;
		}
	}

	ret = PRINT_CLASS (pc)->rgbimage (pc, b, width, height, width * 3);

	g_free (b);

	return ret;
}

int
gnome_print_pixbuf (GnomePrintContext *pc, GdkPixbuf *pixbuf)
{
       if (gdk_pixbuf_get_has_alpha (pixbuf))
               gnome_print_rgbaimage  (pc,
                                       gdk_pixbuf_get_pixels    (pixbuf),
                                       gdk_pixbuf_get_width     (pixbuf),
                                       gdk_pixbuf_get_height    (pixbuf),
                                       gdk_pixbuf_get_rowstride (pixbuf));
       else
               gnome_print_rgbimage  (pc,
                                      gdk_pixbuf_get_pixels    (pixbuf),
                                      gdk_pixbuf_get_width     (pixbuf),
                                      gdk_pixbuf_get_height    (pixbuf),
                                      gdk_pixbuf_get_rowstride (pixbuf));
	return 1;
}

int
gnome_print_textline (GnomePrintContext *pc, GnomeTextLine *line)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (line != NULL, -1);

	return PRINT_CLASS(pc)->textline (pc, line);
}

int
gnome_print_showpage (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->showpage (pc);
}

int
gnome_print_beginpage (GnomePrintContext *pc, const char *name_of_this_page)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->beginpage (pc, name_of_this_page);
}

int
gnome_print_setopacity (GnomePrintContext *pc, double opacity)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	return PRINT_CLASS(pc)->setopacity (pc, opacity);
}

int
gnome_print_context_close (GnomePrintContext *pc)
{
	return PRINT_CLASS(pc)->close (pc);
}



/*
 * These functions are wrapper functions around capabilities that may
 * or may not be deconvolved in back ends.
 */

int
gnome_print_scale (GnomePrintContext *pc, double sx, double sy)
{
	double dst[6];

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	art_affine_scale (dst, sx, sy);
	return gnome_print_concat (pc, dst);
}

int
gnome_print_rotate (GnomePrintContext *pc, double theta)
{
	double dst[6];

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	art_affine_rotate (dst, theta);
	return gnome_print_concat (pc, dst);
}

int
gnome_print_translate (GnomePrintContext *pc, double x, double y)
{
	double dst[6];

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	art_affine_translate (dst, x, y);
	return gnome_print_concat (pc, dst);
}

/*
 * These functions provide a common interface for writing bytes to the
 *
 * printer.
 */

/* Return 0 on success */
int
gnome_print_context_open_file (GnomePrintContext *pc, const char *filename)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (filename != NULL, -1);

	if (filename[0] == '|')
	{
		pc->f = popen (filename + 1, "w");
		pc->is_pipe = TRUE;
	}
	else
	{
		/* This is not very intelligent, but it worked for me */

		if ((filename[0] == '~') && (filename[1] == '/'))
		{
			filename = g_concat_dir_and_file (g_get_home_dir (), &filename[2]);
		}
		pc->f = fopen (filename, "w");
		pc->is_pipe = FALSE;
	}
	return pc->f != NULL;
}

/* Return number of bytes written */
int
gnome_print_context_write_file (GnomePrintContext *pc, char *buf, size_t size)
{
	g_return_val_if_fail (pc != NULL, 0);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), 0);
	g_return_val_if_fail (pc->f != NULL, 0);

	return fwrite (buf, sizeof(char), size, pc->f);
}

/* Return number of bytes written, or -1 if error */
int
gnome_print_context_fprintf (GnomePrintContext *pc, const char *fmt, ...)
{
	va_list ap;
	int n_bytes;
	char *oldlocale;

	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (pc->f != NULL, -1);

	oldlocale = g_strdup (setlocale (LC_NUMERIC, NULL));
	setlocale (LC_NUMERIC, "C");

	va_start (ap, fmt);
	n_bytes = vfprintf (pc->f, fmt, ap);
	va_end (ap);

	setlocale (LC_NUMERIC, oldlocale);
	g_free (oldlocale);

	return n_bytes;
}

/* Return 0 on success */
int
gnome_print_context_close_file (GnomePrintContext *pc)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);

	if (pc->f != NULL)
	{
		if (pc->is_pipe)
			return pclose (pc->f);
		else
			return fclose (pc->f);
		pc->f = NULL;
	}
	else
		return 0;
}

void
gnome_print_bpath (GnomePrintContext * gpc, ArtBpath * bpath, gboolean append)
{
	ArtBpath * p;
	gboolean closed;

	g_return_if_fail (gpc != NULL);
	g_return_if_fail (GNOME_IS_PRINT_CONTEXT (gpc));
	g_return_if_fail (bpath != NULL);

	if (bpath->code == ART_END) return;

	g_return_if_fail ((bpath->code == ART_MOVETO) || (bpath->code == ART_MOVETO_OPEN));

	closed = (bpath->code == ART_MOVETO);

	if (!append)
		gnome_print_newpath (gpc);

	gnome_print_moveto (gpc,
			    bpath->x3,
			    bpath->y3);

	for (p = bpath + 1; p->code != ART_END; p++) {
		switch (p->code) {
		case ART_MOVETO:
		case ART_MOVETO_OPEN:
			if (closed)
				gnome_print_closepath (gpc);

			closed = (p->code == ART_MOVETO);

			gnome_print_moveto (gpc,
					    p->x3,
					    p->y3);
			break;
		case ART_LINETO:
			gnome_print_lineto (gpc,
					    p->x3,
					    p->y3);
			break;
		case ART_CURVETO:
			gnome_print_curveto (gpc,
					     p->x1,
					     p->y1,
					     p->x2,
					     p->y2,
					     p->x3,
					     p->y3);
			break;
		default:
			g_log (G_LOG_DOMAIN,
			       G_LOG_LEVEL_CRITICAL,
			       "Invalid Bpath element");
			return;
			break;
		}
	}

	if (closed)
		gnome_print_closepath (gpc);

}



void
gnome_print_vpath (GnomePrintContext * gpc, ArtVpath * vpath, gboolean append)
{
	ArtVpath * p;
	gboolean closed;

	g_return_if_fail (gpc != NULL);
	g_return_if_fail (GNOME_IS_PRINT_CONTEXT (gpc));
	g_return_if_fail (vpath != NULL);

	if (vpath->code == ART_END) return;

	g_return_if_fail ((vpath->code == ART_MOVETO) || (vpath->code == ART_MOVETO_OPEN));

	closed = (vpath->code == ART_MOVETO);

	if (!append)
		gnome_print_newpath (gpc);

	gnome_print_moveto (gpc,
			    vpath->x,
			    vpath->y);

	for (p = vpath + 1; p->code != ART_END; p++) {
		switch (p->code) {
		case ART_MOVETO:
		case ART_MOVETO_OPEN:
			if (closed)
				gnome_print_closepath (gpc);

			closed = (p->code == ART_MOVETO);

			gnome_print_moveto (gpc,
					    p->x,
					    p->y);
			break;
		case ART_LINETO:
			gnome_print_lineto (gpc,
					    p->x,
					    p->y);
			break;
		default:
			g_log (G_LOG_DOMAIN,
			       G_LOG_LEVEL_CRITICAL,
			       "Invalid Vpath element");
			return;
			break;
		}
	}

	if (closed)
		gnome_print_closepath (gpc);

}

/*
 * WARNING: EXPERIMENTAL
 */

int
gnome_print_glyphlist (GnomePrintContext * pc, GnomeGlyphList * glyphlist)
{
	g_return_val_if_fail (pc != NULL, -1);
	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (pc), -1);
	g_return_val_if_fail (glyphlist != NULL, -1);
	g_return_val_if_fail (GNOME_IS_GLYPHLIST (glyphlist), -1);

	if (((GnomePrintContextClass *) ((GtkObject *) pc)->klass)->glyphlist) {
		return (* ((GnomePrintContextClass *) ((GtkObject *) pc)->klass)->glyphlist) (pc, glyphlist);
	} else {
		GnomePosGlyphList * pgl;
		gdouble affine[6];
		GSList * l;
		art_affine_identity (affine);
		pgl = gnome_pgl_from_gl (glyphlist, affine, 0);
		for (l = pgl->strings; l != NULL; l = l->next) {
			GnomePosString *string;
			GnomeFont *font;
			GnomeFontFace *face;
			gint i;
			string = (GnomePosString *) l->data;
			font = (GnomeFont *) gnome_rfont_get_font (string->rfont);
			face = (GnomeFontFace *) gnome_font_get_face (font);
			gnome_print_setfont (pc, font);
			for (i = 0; i < string->length; i++) {
				gdouble r, g, b, a;
				const gchar *psname;
				gint ucschar;
				/* fixme: We need currentpoint here :( */
				gnome_print_moveto (pc, string->glyphs[i].x, string->glyphs[i].y);
				r = ((string->glyphs[i].color >> 24) & 0xff) / 255.0;
				g = ((string->glyphs[i].color >> 16) & 0xff) / 255.0;
				b = ((string->glyphs[i].color >>  8) & 0xff) / 255.0;
				a = ((string->glyphs[i].color >>  0) & 0xff) / 255.0;
				gnome_print_setrgbcolor (pc, r, g, b);
				gnome_print_setopacity (pc, a);
				psname = gnome_font_face_get_glyph_ps_name (face, string->glyphs[i].glyph);
				ucschar = GINT_TO_BE (gp_unicode_from_ps (psname));
				gnome_print_show_ucs4 (pc, &ucschar, 1);
			}
		}
		gnome_pgl_destroy (pgl);
	}

	return 0;
}











